/* N-body calculation using Greenspan discrete mechanics.		*/
/* Uses 3*N*N (of 4*N*N) processors, communication by spreading and	*/
/* reduction.  Incorporates xyz-parallelism, all processors store	*/
/* particle information and calculate velocities, positions.		*/
/* File:  nb3ds2.cs							*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <cscomm.h>

/* Max number of bodies in the calculation.  N = 64 fits in 16k procs	*/
/* N must be a power of 2, 4N^2 >= physical CM size			*/

#define N 64

int   NB;	/* Number of bodies in system, <= N	*/
int   PX;	/* Controls how much state is printed	*/
float G;	/* Gravitational constant		*/
float MINF;	/* Minimum CM single-float		*/

#define TWOPI      6.2831853f

/* Particles, and pairs of particles.  Diagonal entries have particle	*/
/* states.  Some variables could be overloaded. 			*/

shape [4][N][N]pair;

/* x0, x1 - current, new positions;	v0, v1 - current, new velocities */
/* m - mass;		mj - scaled mass;	ac - acceleration	 */
/* w, ax - working	r0 - distance (X0)	d0 - displacement (X0)	 */

float:pair x0, x1, v0, v1, m, mj, ac, d0, r0, ax, w;

/* i - row number;    j - column number;    c - coordinate number	*/

int:pair i, j, c;

/* sm   - processor is part of computation = diag | act	*/
/* diag - processor is on diagonal			*/
/* act  - processor is active pair (off-diag)		*/

bool:pair sm, diag, act;

/* Pairs are represented symmetrically.	 Setup initializes these.	*/

void setup()
 { with (pair)
    { /* Initialize coordinates */
      c = pcoord(0);  i = pcoord(1);  j = pcoord(2);
      /* Identify active processors */
      sm = (i < NB) & (j < NB) & c < 3;
      act = sm & (i != j);  diag = sm & (i == j);
    }
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */
/* Leaves data for Pi in processors [c][i][*].				*/

void init()
 { int k;  float a, b, c;
   with (pair)
   {for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);
       [0][k][0]x0 = a;  [1][k][0]x0 = b;  [2][k][0]x0 = c;
     };
    for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);
       [0][k][0]v0 = a;  [1][k][0]v0 = b;  [2][k][0]v0 = c;
     };
    for ( k = 0; k < NB; k++ )
      { fscanf(datap, "%e", &a);
        [0][k][0]m = [1][k][0]m = [2][k][0]m = a;
      };
    fscanf(datap, "%e", &G);
    /* Spread data across rows */
    x0 = copy_spread(&x0, 2, 0);
    v0 = copy_spread(&v0, 2, 0);
    m  = copy_spread(&m,  2, 0);
   }
 }

/* Insert nx asteroids at radius ar with total mass am.  This assumes	*/
/* that an orbit of radius 1 about origin corresponds to velocity 2*pi	*/
/* (e.g., au-earth-yr units).  Uses parallel trig.			*/

void mkast(int nx, float ar, float am)
 { float:pair sx, cx, arg;
   float spd;  int nn;
   nn = NB + nx;
   spd = TWOPI/sqrt(ar);
   with (pair) 
   { sm = (i < nn) & (j < nn) & c < 3;
     act = sm & (i != j);  diag = sm & (i == j);
     where ( (NB <= i) & (i < nn) & sm )
      { arg = (TWOPI*(i - NB))/nx;
        sx = sin(arg);  cx = cos(arg);
        where ( c == 0 ) { x0 =  cx;  v0 = -sx; }
        where ( c == 1 ) { x0 =  sx;  v0 =  cx; }
        where ( c == 2 ) { x0 = 0.0;  v0 = 0.0; }
        x0 *= ar;  v0 *= spd;  m = am/nx;
      }}  /* end where, end with (pair) */
   NB = nn;
 }

/* Print current state of system according to PX.  See comments for	*/
/* main program for PX settings.					*/

void print_state()
 { int k, np;
   if ( (PX >= 0) && (PX < NB) ) np = PX; else np = NB;
   printf(
"\n       x         y         z         vx        vy        vz        m\n"
         );
   for ( k = 0; k < np; k++ )
     printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
            k, [0][k][k]x0, [1][k][k]x0, [2][k][k]x0,
               [0][k][k]v0, [1][k][k]v0, [2][k][k]v0, [0][k][k]m);
 }

/* Print x-components of xx in nx by nx processors (for testing). */

/*********************************************************************

void print_x(char lbl[], float:current xx, int nx)
 { int u, v;
   printf("\n%s\n", lbl); 
   for ( u = 0; u < nx; u++)
    { for ( v = 0;  v < nx; v++)
        printf("%9.2e ", [0][u][v]xx);
      printf("\n");
    }
 }

*********************************************************************/

/* Calculate accelerations.  r0, d0 already give displacement, distance  */
/* for current position.  Called with all processors active.		 */

void accel()
 { float:pair d1, r1;
   ac = 0.0;  ax = MINF;		/* MINF is most neg. float */
   where (diag) ax = x1;		/* Prepare to spread */
   d1 = spread(ax, 1, CMC_combiner_max);
   /* print_x("accel, d1 after spread 1", d1, NB);	 For testing */
   where (act)
    { /* Now d1 = [c][j][j]x1,  x1 = [c][i][i]x1 */      
      d1 = d1 - x1;			/* New displacement component */
      r1 = d1*d1;			/* For sum of squares */
      w  = spread(r1, 0, CMC_combiner_add);
      r1 = sqrt(w);			/* Euclidean distance */
      /* print_x("accel, r1", r1, NB);	 For testing */
      ac = mj*(d0+d1)/(r0*r1*(r0+r1));	    /* Acceleration component */
      /* print_x("accel, initial ac", ac, NB);	 For testing */
    } /* end where (act) */
   ac = spread(ac, 2, CMC_combiner_add); 	 /* Sum accelerations */
   /* print_x("accel, final ac", ac, NB);		 For testing */
  }

/* Calculate initial accelerations.  Called with all proc's active. */

void accel1()
 { ac = 0.0;  ax = MINF;		/* MINF is most neg. float */
   where (diag) ax = x0;		/* Prepare to spread */
   d0 = spread(ax, 1, CMC_combiner_max);
   /* print_x("accel1, d0 after spread 1", d0, NB);	 For testing */
   where (act)
    { /* Now d0 = [c][j][j]x0,  x0 = [c][i][i]x0 */
      d0 = d0 - x0;			/* New displacement component */
      r0 = d0*d0;			/* For sum of squares */
      w  = spread(r0, 0, CMC_combiner_add);
      r0 = sqrt(w);			/* Euclidean distance */
      /* print_x("accel1, r0", r0, NB);	 For testing */
      ac = mj*d0/(r0*w);		/* Acceleration component */
      /* print_x("accel1, initial ac", ac, NB);	 For testing */
    } /* end where (act) */
   ac = spread(ac, 2, CMC_combiner_add);        /* Sum accelerations */
 }

/* Greenspan's discrete mechanics.  Performs ns steps of time dt	*/
/* with solution tolerance eps.  Returns number of evaluations.		*/
/* Leaves number of iterations which were stopped at limit in cfail.	*/

int cfail;   /* Counts number of iterations that failed to converge */

int green1(int ns, float dt, float eps)
 { int k, ic, v, lim=10;  float hdt, mx, my;
   v = 0;  hdt = 0.5*dt;
   with (pair)
   { float:pair nx;
     for (k = 0; k < ns; k++)
       { ic = 0;  accel1();
	 /* This is a do - while - repeat loop: */
   	 do
          { v++;  ic++;
	    where (sm)
	    { nx = v0 + dt*ac;			/* New velocity	*/
	      mx = (>?= fabs(nx - v1));		/* Max change	*/
	      v1 = nx;
   	      /* print_x("green1, v1", v1, NB);	 For testing */
	      nx = x0 + hdt*(v0 + v1);		/* New position	*/
	      my = (>?= fabs(nx - x1));		/* Max change	*/
	      if (my > mx) mx = my;		/* Max change x, v */
   	      /* print_x("green1, x1", nx, NB);	 For testing */
	      x1 = nx; }
	    if ( (mx < eps) || ( ic == lim ) ) break;
	    accel(); }
	 while (1);
         where (sm) { x0 = x1;  v0 = v1; };  if (ic == lim) cfail++;
       } ;   /* end for */
    }  /* end  with (pair) */
   return(v);
 }

/* Main program:  reads problem and control information from data file	*/
/* named by user and performs calculation.				*/
/* Print control:							*/
/*	  0 - just performance data					*/
/*	< 0 - all							*/
/*	> 0 - number given						*/

void main()
 { int ns, np, k, vl;  float dt, eps, ar, am;
   char fnm[50];
   /* Initialize MINF = min CM float = max over empty set */
   with (pair)
    { sm = 0;  x0 = 0.0;  where (sm) MINF = (>?= x0); }
   /* Loop for test cases (tbs) */
   while ( printf("Data file (or q)> "),
	   scanf("%s", fnm),
	   ( fnm[0] != 'q' ) || ( fnm[1] > 0 ) )
    { printf("Print ctl> ");  scanf("%d", &PX);
      /* Performs np integrations, each ns steps of dt */
      datap = fopen(fnm, "r");
      fscanf(datap, "%d", &NB);
      setup();   /* Setup geometric structure */
      init();    /* Read data for individual bodies */
      /* Read "asteroid" descriptions */
      while ( fscanf(datap, "%d", &ns), ns > 0)
       { fscanf(datap, "%e%e", &ar, &am);  mkast(ns, ar, am); }
      fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);
      fclose(datap);
      printf("\n   NB = %d", NB);
      printf("\n   G = %13.6e  delta-t = %12.5e  epsilon = %12.5e\n",
                    G, dt, eps);
      printf("\n simulated time = 0");
      print_state();
      /* Spread masses and scale */
      with (pair) where (act) { m = [c][j][j]m;  mj = G*m; }
      for ( k = 0; k < np; k++ )
       { CM_timer_clear(1);  CM_timer_start(1);
         cfail = 0;
         vl = green1(ns, dt, eps);
         CM_timer_stop(1);  CM_timer_print(1);
         printf(
     "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
	          (k + 1)*ns*dt, vl,            cfail);
         print_state();
       };  /* end for */
    }  /* end while */
 }

/*********************************************************************
  Data files:  as for nb3dg2.cs
*************************************************************************/
